#include "public.h"
#include "common.h"
#include "g72x.h"
#include "audio.h"

// include files for socket communication
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <cstring>
#include <errno.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/socket.h>

unsigned char orgBuf[1040000];
int strmLen;

bool play(AUDIOFMT* decodeBuf)
{
    if (write(audio_fd, (char*)&decodeBuf[0], BUF_SIZE_BYTE)!= BUF_SIZE_BYTE)
        return false;
            
    return true;
}

int unpack_input(unsigned char *code, int bits, unsigned char* encodeBuf)
{   
    static unsigned int in_buffer = 0;
    static int      in_bits = 0;
    unsigned char       in_byte;

    static int i=0;

    if (in_bits < bits)
    {   
        in_byte=encodeBuf[i];
        i=(i+1)%COMP_SIZE;
        in_buffer |= (in_byte << in_bits);
        in_bits += 8;
    }
    *code = in_buffer & ((1 << bits) - 1);
    in_buffer >>= bits;
    in_bits -= bits;
    return (in_bits > 0);
}


void decode(AUDIOFMT* decodeBuf, unsigned char* encodeBuf)
{
    unsigned char       code;
    signed short sample;

    int i;

    static int decodeFirst=1;
    static struct g72x_state  destate;

    if (decodeFirst==1)
    {
 	// This segment should be called for the first frame only.
	// Subsequent frames should bypass this part of the loop.
        g72x_init_state(&destate);
        decodeFirst=0;
    }

    for (i=0;i<BUF_SIZE;i++)
    {
        unpack_input(&code,4, encodeBuf);
        //decodeBuf[i]=(short)g721_decoder(code,AUDIO_ENCODING_LINEAR,&destate);
        decodeBuf[i]=(short)g721_decoder(code,AUDIO_ENCODING_ULAW,&destate);
    }

}

void read_file() 
{ 
  FILE *fstart = fopen("orgstream.count", "rb");
  assert(fstart != NULL);
  strmLen = 0;
  
  while(fread(orgBuf + strmLen, sizeof(char), BUF_SIZE, fstart) == BUF_SIZE) {
    strmLen += BUF_SIZE;
  }
}
  
  
void main(int argc, char *argv[])
{
    AUDIOFMT decodeBuf[BUF_SIZE];
    unsigned char encodeBuf[COMP_SIZE];
    //   unsigned char orgBuf[BUF_SIZE];
    unsigned char intlvDecodeBuf[INTLV_DEGREE][INTLV_BUF_SIZE];
    bool recvd[INTLV_DEGREE];
    unsigned char buffer[sizeof(Pkt)];
    unsigned char accum[sizeof(Pkt)];
    Pkt pckt, pckt1;
    unsigned int len;
    int curr_tuple = 0;
    int last_tupid = 0;
    int tupid, streamid;
    int datalen, ret;

    read_file();

    double num=0.0, den=0.0, snr=0.0;
    
    //initialize the socket to hear on...
    struct hostent *he;
    int s, conn_s, lis_s;
    struct sockaddr_in ls;
    
    //initialize the TCP socket to get data from recver
    if((s = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
      perror("socket creation failed");
      exit(1);
    }

    bzero(&ls, sizeof(struct sockaddr_in));
   
    ls.sin_family = AF_INET;
    ls.sin_port = htons(pPORT);
    inet_aton(ADDR_P, (struct in_addr *)&ls.sin_addr);

    //connect to the local receiver
    if(bind(s, (struct sockaddr *)&ls, sizeof(struct sockaddr_in)) == -1) {
      perror("bind recv failed: ");
      exit(1);
    }
    
    //listen and accept a connect
    if(listen(s, 5) == -1){
      perror("listen recv: ");
      exit(2);
    }
    int ln=sizeof(struct sockaddr_in);
    if((conn_s=accept(s, (struct sockaddr *)&ls, (socklen_t *)&ln)) == -1){
      perror("accept recv: ");
      exit(3);
    }

    //read off conn_s
    
    //open_audio();

    int numpkts=0, numlosses = 0;
    struct timeval tm;
    int sec=0, usec=0;
    int sec1=0, usec1=0;
    double jitter=0.0;
    double delay=0.0, curr_delay=0.0;
    double avg_lateness = 0.0;

    for(int i=0;i<INTLV_DEGREE;i++)
      recvd[i] = false;

    while(1)
      {
	datalen = sizeof(Pkt);
	while(datalen > 0) {
	  if((ret = recv(conn_s, buffer, datalen, 0)) == -1) {
	    perror("recv");
	    exit(1);
	  }
	  bcopy(buffer, &accum[sizeof(Pkt) - datalen], ret);
	  datalen -= ret;
	}
	
	// one packet complete
	bcopy(accum, &pckt, sizeof(Pkt));
		
	gettimeofday(&tm, NULL);

	if (sec != 0) { // this is not the first packet
	  // we have previous values recorded
	  int dif = pckt.tupleid-last_tupid;
	  double curr_delay = (tm.tv_sec -sec) + (tm.tv_usec -usec)/1000000.0;
	  curr_delay = curr_delay/dif;
		
	  if(delay !=0){
	    double dd = (curr_delay - delay);
	    if(dd < 0)
		dd = 0;
			      
	    jitter = (numpkts*jitter + dd)/(numpkts+1);
	  }
	  delay = ((numpkts)*delay + curr_delay)/(numpkts+1);
	}
	sec = tm.tv_sec;
	usec = tm.tv_usec;

	last_tupid = pckt.tupleid;
	numpkts++;

	tupid = (pckt.tupleid)/INTLV_DEGREE;
	streamid = (pckt.tupleid)%INTLV_DEGREE;
	if(pckt.tupleid >= NUM_PCKTS) {
	  close(s);
		printf("Exiting gracefully\n");
	  exit(0);
	}
	
	//printf("Received a packet: %d, %d\n", tupid, streamid);
	
	if(tupid == curr_tuple) {
	  recvd[streamid] = true;
	  decode(intlvDecodeBuf[streamid], pckt.data);
	}
	else {
	  // either the current tuple is full or something went amiss
	  // reconstruct the current tuple (if needed) and play it
	  int x, y1, y2;
	  // using linear interpolation
	  for(int i=0; i<INTLV_DEGREE; i++) {
	    if(recvd[i] == false) {
	      //printf("\t \n\nsome lost tuples.\n\n\n");
	      // a missing sample
	      // try to find the earlier adjacent sample
	      int j,k,k1,k2,l;
	      bool flag = false;
	      x = y1 = y2  = 0;
	      for(j=1; j<INTLV_DEGREE;j++) {
		l = i-j;
		if(l<0) {
		  x = -1;
		  l = INTLV_DEGREE + l;
		}
		if(recvd[l]) {
		  flag = true;
		  break;
		}
	      }
	      if(!flag) {
		// no packets in stream!
		// feed silence into everything
		for(j=0; j<INTLV_DEGREE; j++) 
		  bzero(intlvDecodeBuf[j], INTLV_BUF_SIZE);
		break;
	      }
	      j = l;
	      flag=false;
	      for(k=1; k<2*INTLV_DEGREE; k++) {
		l = i + k;
		if(!flag && l>=INTLV_DEGREE) {
		  y1 = 1;
		  l = l - INTLV_DEGREE;
		}
		if(flag && l>=INTLV_DEGREE && l<2*INTLV_DEGREE) {
		  y2=1;
		  l = l - INTLV_DEGREE;
		}
		if(flag && l>=2*INTLV_DEGREE) {
		  y2=2;
		  l = l - (2*INTLV_DEGREE);
		}
		if(recvd[l] && flag)
		  break;
		if(recvd[l]) {
		  k1 = l;
		  flag = true;
		}
	      }
	      k2=l;
	      // now reconstruct the missing parts
	      int lhs, rhs, rrhs;
	      for(int m=0; m < INTLV_BUF_SIZE; m++) {
		if(m+x < 0) {
		  intlvDecodeBuf[i][m] = intlvDecodeBuf[k1][m]; // y wud be zero
		  continue;
		}
		else lhs = m+x;
		if(m+y1 >= INTLV_BUF_SIZE) {
		  intlvDecodeBuf[i][m] = intlvDecodeBuf[j][m]; // x wud be zero
		  continue;
		}
		else {
		  rhs = m + y1;
		  if(m+y2 >= INTLV_BUF_SIZE) {
		    intlvDecodeBuf[i][m] = (intlvDecodeBuf[j][lhs] + intlvDecodeBuf[k1][rhs])/2;
		    continue;
		  }
		  else
		    rrhs = m + y2;
		}
		intlvDecodeBuf[i][m] = intlvDecodeBuf[j][lhs] + (intlvDecodeBuf[k1][rhs] - intlvDecodeBuf[j][lhs])/2 - (intlvDecodeBuf[j][lhs] - 2*intlvDecodeBuf[k1][rhs] + intlvDecodeBuf[k2][rrhs])/8;
	      }
	    }
	  }
	  // now the entire decode buf is reconstructed, let's play
	  
	  int ptr=0;
	  int i;
	  while(ptr < BUF_SIZE)
	    for(i=0; i<INTLV_DEGREE; i++, ptr++)
	      decodeBuf[ptr] = intlvDecodeBuf[i][ptr/INTLV_DEGREE];
	  
	  //snr...
	  int pv = (BUF_SIZE * curr_tuple)%strmLen;
	  
	  for(int z=0; z<BUF_SIZE; z++) {
	    num += (orgBuf[pv + z]*orgBuf[pv + z]);
	    den += (orgBuf[pv + z] - decodeBuf[z])*(orgBuf[pv + z] - decodeBuf[z]);
	  }
	  snr = 10*(log10(num/den));
	  
	    //printf("tuple  = %d, jitter=%lf, delay=%lf, snr = %lf, numlosses=%d\n", curr_tuple, jitter, delay, snr, numlosses);	  
	    printf("%d %lf %lf %lf %d\n", curr_tuple, jitter, delay, snr, numlosses);	  
	    fflush(stdout);
	  
	  //	if (!play(decodeBuf))
	  // break;
	  
	  curr_tuple++;
	  while(curr_tuple != tupid) {
	    ++numlosses;
	    //printf("\n\n\nWhole tuple lost: %d\n", (++numlosses));
	    bzero(decodeBuf, BUF_SIZE);

	    pv = (BUF_SIZE * curr_tuple)%strmLen;
	    for(int z=0; z<BUF_SIZE; z++) {
	      num += (orgBuf[pv + z]*orgBuf[pv + z]);
	      den += (orgBuf[pv + z] - decodeBuf[z])*(orgBuf[pv + z] - decodeBuf[z]);
	    }
	    snr = 10*(log10(num/den));
	    
	    //printf("tuple  = %d, jitter=%lf, delay=%lf, snr = %lf, numlosses=%d\n", curr_tuple, jitter, delay, snr, numlosses);	  
	    printf("%d %lf %lf %lf %d\n", curr_tuple, jitter, delay, snr, numlosses);	  
	    fflush(stdout);
	    // if (!play(decodeBuf))
	    //  break;
	    curr_tuple++;
	  }
	  for(int i=0;i<INTLV_DEGREE;i++)
	    recvd[i] = false;
	  
	  recvd[streamid] = true;
	  decode(intlvDecodeBuf[streamid], pckt.data);
      }
/* 
	if (!play(decodeBuf))
	break;
*/
    }
    snr = 10*(log10(num/den));

    printf("%lf\n", snr);
    //close_audio();
}
