#include "public.h"
#include "common.h"
#include "g72x.h"
#include "audio.h"

// include files for socket communication
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <cstring>
#include <errno.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/socket.h>

bool play(AUDIOFMT* decodeBuf)
{
    if (write(audio_fd, (char*)&decodeBuf[0], BUF_SIZE_BYTE)!= BUF_SIZE_BYTE)
        return false;
            
    return true;
}

int unpack_input(unsigned char *code, int bits, unsigned char* encodeBuf)
{   
    static unsigned int in_buffer = 0;
    static int      in_bits = 0;
    unsigned char       in_byte;

    static int i=0;

    if (in_bits < bits)
    {   
        in_byte=encodeBuf[i];
        i=(i+1)%INTLV_COMP_SIZE;
        in_buffer |= (in_byte << in_bits);
        in_bits += 8;
    }
    *code = in_buffer & ((1 << bits) - 1);
    in_buffer >>= bits;
    in_bits -= bits;
    return (in_bits > 0);
}


void decode(AUDIOFMT* decodeBuf, unsigned char* encodeBuf)
{
    unsigned char       code;
    signed short sample;

    int i;

    static int decodeFirst=1;
    static struct g72x_state  destate;

    if (decodeFirst==1)
    {
 	// This segment should be called for the first frame only.
	// Subsequent frames should bypass this part of the loop.
        g72x_init_state(&destate);
        decodeFirst=0;
    }

    for (i=0;i<INTLV_BUF_SIZE;i++)
    {
        unpack_input(&code,4, encodeBuf);
        //decodeBuf[i]=(short)g721_decoder(code,AUDIO_ENCODING_LINEAR,&destate);
        decodeBuf[i]=(short)g721_decoder(code,AUDIO_ENCODING_ULAW,&destate);
    }

}

void main(int argc, char *argv[])
{
    AUDIOFMT decodeBuf[BUF_SIZE];
    unsigned char encodeBuf[COMP_SIZE];
    unsigned char intlvDecodeBuf[INTLV_DEGREE][INTLV_BUF_SIZE];
    bool recvd[INTLV_DEGREE];
    unsigned char buffer[sizeof(Pkt)];
    unsigned char accum[sizeof(Pkt)];
    Pkt pckt, pckt1;
    FILE* fpen;
    FILE *fpen1;
    unsigned int len;
    int curr_tuple=0;
    int tupid, streamid;
    int datalen, ret;
    int s, lis_s;
    struct sockaddr_in rs; 
    struct sockaddr_in ls;

    // initialize the recvd bit vector
    for(int i=0;i<INTLV_DEGREE;i++)
      recvd[i] = false;
    
    /* create a server (listening) socket */
    if((lis_s = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
      perror("listening socket");
      exit(1);
    }

    bzero(&rs, sizeof(struct sockaddr_in));
    bzero(&ls, sizeof(struct sockaddr_in));

    ls.sin_family = AF_INET;
    ls.sin_port = htons(atoi(argv[1]));
    inet_aton(RADDR, (struct in_addr *)&ls.sin_addr);

    if(bind(lis_s, (struct sockaddr *)&ls, sizeof(struct sockaddr_in)) == -1) {
      perror("bind");
      exit(1);
    }
    
    if(listen(lis_s, 5) == -1) {
      perror("listen");
      exit(1);
    }
    len = sizeof(struct sockaddr_in);
    
    if((s = accept(lis_s, (struct sockaddr *)&rs, &len)) == -1) {
      perror("accept");
      exit(1);
    }
    
    //FILE *fpen6 = fopen("./codedstream", "rb");
    while(1)
    {
      //I read in the encoded stream from the file
      //You can read it in from the network

      /* read the stream from the network */
      
      
      datalen = sizeof(Pkt);
      while(datalen > 0) {
	if((ret = recv(s, buffer, datalen, 0)) == -1) {
	  perror("recv");
	  exit(1);
	}
	bcopy(buffer, &accum[sizeof(Pkt) - datalen], ret);
	datalen -= ret;
      }
      
      // one packet complete
      bcopy(accum, &pckt, sizeof(Pkt));
      
      
      /*
	if(fread(&pckt, sizeof(char), sizeof(Pkt), fpen6)<sizeof(Pkt))
	break;
      */

      tupid = (pckt.tupleid)/INTLV_DEGREE;
      streamid = (pckt.tupleid)%INTLV_DEGREE;

      if(pckt.tupleid >= NUM_PCKTS) {
	fclose(fpen1);
	close(s);
	close(lis_s);
	exit(0);
      }

      printf("Received a packet: %d, %d\n", tupid, streamid);

      if(tupid == curr_tuple) {
	recvd[streamid] = true;
	decode(intlvDecodeBuf[streamid], pckt.data);
      }
      else {
	// either the current tuple is full or something went amiss
	// reconstruct the current tuple (if needed) and play it
	int x, y1, y2;
	// using linear interpolation
	for(int i=0; i<INTLV_DEGREE; i++) {
	  if(recvd[i] == false) {
	    printf("\t some lost tuples.\n");
	    // a missing sample
	    // try to find the earlier adjacent sample
	    int j,k,k1,k2,l;
	    bool flag = false;
	    x = y1 = y2  = 0;
	    for(j=1; j<INTLV_DEGREE;j++) {
	      l = i-j;
	      if(l<0) {
		x = -1;
		l = INTLV_DEGREE + l;
	      }
	      if(recvd[l]) {
		flag = true;
		break;
	      }
	    }
	    if(!flag) {
	      // no packets in stream!
	      // feed silence into everything
	      for(j=0; j<INTLV_DEGREE; j++) 
		bzero(intlvDecodeBuf[j], INTLV_BUF_SIZE);
	      break;
	    }
	    j = l;
	    flag=false;
	    for(k=1; k<2*INTLV_DEGREE; k++) {
	      l = i + k;
	      if(!flag && l>=INTLV_DEGREE) {
		y1 = 1;
		l = l - INTLV_DEGREE;
	      }
	      if(flag && l>=INTLV_DEGREE && l<2*INTLV_DEGREE) {
		y2=1;
		l = l - INTLV_DEGREE;
	      }
	      if(flag && l>=2*INTLV_DEGREE) {
		y2=2;
		l = l - (2*INTLV_DEGREE);
	      }
	      if(recvd[l] && flag)
		break;
	      if(recvd[l]) {
		k1 = l;
		flag = true;
	      }
	    }
	    k2=l;
	    // now reconstruct the missing parts
	    int lhs, rhs, rrhs;
	    for(int m=0; m < INTLV_BUF_SIZE; m++) {
	      if(m+x < 0) {
		intlvDecodeBuf[i][m] = intlvDecodeBuf[k1][m]; // y wud be zero
		continue;
	      }
	      else lhs = m+x;
	      if(m+y1 >= INTLV_BUF_SIZE) {
		intlvDecodeBuf[i][m] = intlvDecodeBuf[j][m]; // x wud be zero
		continue;
	      }
	      else {
		rhs = m + y1;
		if(m+y2 >= INTLV_BUF_SIZE) {
		  intlvDecodeBuf[i][m] = (intlvDecodeBuf[j][lhs] + intlvDecodeBuf[k1][rhs])/2;
		  continue;
		}
		else
		  rrhs = m + y2;
	      }
	      intlvDecodeBuf[i][m] = intlvDecodeBuf[j][lhs] + (intlvDecodeBuf[k1][rhs] - intlvDecodeBuf[j][lhs])/2 - (intlvDecodeBuf[j][lhs] - 2*intlvDecodeBuf[k1][rhs] + intlvDecodeBuf[k2][rrhs])/8;
	    }
	  }
	}
	// now the entire decode buf is reconstructed, let's play

	int ptr=0;
	while(ptr < BUF_SIZE)
	  for(int i=0; i<INTLV_DEGREE; i++, ptr++)
	    decodeBuf[ptr] = intlvDecodeBuf[i][ptr/INTLV_DEGREE];

	fwrite(decodeBuf, sizeof(char), BUF_SIZE, fpen1);
	fflush(fpen1);
	
	//	if (!play(decodeBuf))
	// break;
	
	curr_tuple++;
	while(curr_tuple != tupid) {
	  printf("\nWhole tuple lost.\n");
	  bzero(decodeBuf, BUF_SIZE);
	  fwrite(decodeBuf, sizeof(char), BUF_SIZE, fpen1);
	  fflush(fpen1);
	  
	  // if (!play(decodeBuf))
	  //  break;
	  curr_tuple++;
	}
	for(int i=0;i<INTLV_DEGREE;i++)
	  recvd[i] = false;
	
	recvd[streamid] = true;
	decode(intlvDecodeBuf[streamid], pckt.data);
      }
    }
    
    fclose(fpen1);
    //fclose(fpen6);
    close_audio();
}
