#include "public.h"
#include "common.h"
#include "g72x.h"
#include "audio.h"

// include files for socket communication
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <cstring>
#include <errno.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/socket.h>


extern int errno;

//global socket identifiers
int s;

bool record(AUDIOFMT* audioBuf)
{
  int readNum;
  
  readNum=read(audio_fd, (char*)audioBuf, BUF_SIZE_BYTE);
  
  if (readNum!=BUF_SIZE_BYTE)
    return false;
  
  return true;
}

int pack_output(unsigned code,int bits, unsigned char* encodeBuf)
{
  static unsigned int out_buffer = 0;
  static int      out_bits = 0;
  unsigned char   out_byte;
  
  static int i=0;
  
  out_buffer |= (code << out_bits);
  out_bits += bits;
  if (out_bits >= 8)
    {
      out_byte = out_buffer & 0xff;
      out_bits -= 8;
      out_buffer >>= 8;
      encodeBuf[i]=out_byte;
      i=(i+1)%INTLV_COMP_SIZE;
    }
  return (out_bits > 0);
}


void encode(AUDIOFMT* audioBuf, unsigned char* encodeBuf)
{
  unsigned char       code;
  int         resid;
  
  int i;
  
  static int encodeFirst=1;
  static struct g72x_state  enstate;
    
  if (encodeFirst==1)
    {
      // This segment should be called for the first frame only.
      // Subsequent frames should bypass this part of the loop.
      g72x_init_state(&enstate);
      encodeFirst=0;
    }   
  
  for (i=0;i<INTLV_BUF_SIZE;i++)
    {
      //code=g721_encoder(audioBuf[i],AUDIO_ENCODING_LINEAR, &enstate);
      code=g721_encoder(audioBuf[i],AUDIO_ENCODING_ULAW, &enstate);
      resid=pack_output(code,4, encodeBuf);
    }   
  while (resid)
    {
      assert(false);
      resid = pack_output(0, 4, encodeBuf);
    }     
}

void my_fin(int a){
  printf("SIGINT trapped");
  close(s);
  exit(10);
}

void main(int argc, char *argv[])
{
  AUDIOFMT intLvBuf[INTLV_DEGREE][INTLV_BUF_SIZE];
  AUDIOFMT audioBuf[BUF_SIZE];
  unsigned char encodeBuf[INTLV_DEGREE][INTLV_COMP_SIZE];
  unsigned int id=0;
  struct timeval tm;
  unsigned int sec = 0, usec = 0;
  double delay = 0.0;
  char src[1040000];

  Pkt pckt, pckt1;

  FILE *freader;
  
  
  signal(SIGINT, my_fin);
  
  //read the total file
  freader=fopen("orgstream.count", "rb");
  assert(freader!=NULL);
  int fileind=0, filesize=0;
  int tmp=0;
  while((tmp=fread(src+fileind, sizeof(char), BUF_SIZE, freader)) == BUF_SIZE)
    fileind+=tmp;
  fileind+=tmp;

  printf("Read in %d bytes.\n", fileind);
  filesize=fileind;
  fileind=0;
  fclose(freader);//done


  //create the socket
  struct hostent *he;
  int datalen;
  struct sockaddr_in rs, ls; 
  
  //initialize the TCP socket to send to the remote TCP Echo port
  if((s = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    perror("socket creation failed");
    exit(1);
  }

  bzero(&rs, sizeof(struct sockaddr_in));
  bzero(&ls, sizeof(struct sockaddr_in));
  
  if(argc<2){
    printf("Command line argument for remote udp echo server needed\n");
    exit(3);
  }

    // get the address of the remote tcp echo server from the name/IP
    // specified on the command line
    if((he = gethostbyname(argv[1])) == NULL) {
      perror("gethostbyname failed");
      exit(1);
    } 
    
    // fill in the information about the receiver address in the sockaddr_in
    rs.sin_family = AF_INET;
    rs.sin_port =  htons(rPORT); //receiver
    inet_aton(ADDR_SnR, (struct in_addr*)&rs.sin_addr);

    ls.sin_family = AF_INET;
    ls.sin_port =  htons(sPORT); //receiver
    inet_aton(ADDR_SnR, (struct in_addr*)&ls.sin_addr);
    
    if(bind(s, (struct sockaddr *)&ls, sizeof(struct sockaddr_in)) == -1){
      perror ("bind: ");
      exit(1);
    }
    
    //setsockopt for the remote server as argv[1]
    // set TCP_ECHO option
    if(setsockopt(s, 6, 13, he->h_addr, he->h_length) == -1) {
      perror("setsockopt failed");
      exit(1);
    }
   
   int no_nagle=1; 
    if(setsockopt(s, 6, 1, &no_nagle, 4) == -1) {
      perror("setsockopt failed");
      exit(1);
    }
   
    int rtm_snd_size = SND_SIZE;
    if(setsockopt(s, 6, 16, &rtm_snd_size, 4) == -1) {
      perror("setsockopt failed");
      exit(1);
    }
    int rtm_snd_tm = SND_TM;
    if(setsockopt(s, 6, 17, &rtm_snd_tm, 4) == -1) {
      perror("setsockopt failed");
      exit(1);
    }    

    // set maximum segment size
    int maxseg=ECHO_MSS;
    if(setsockopt(s, 6, 2, (char *)&maxseg, sizeof(unsigned long)) == -1) {
      perror("setsockopt failed");
      exit(1);
    }
    //connect to the local receiver
    if(connect(s, (struct sockaddr *)&rs, sizeof(struct sockaddr_in)) == -1) {
      perror("connect local failed: ");
      exit(1);
    }
        
    printf("Starting data transmission\n");
    //single thread that does the sending job
    
    //open_audio();

    int num=0;
    while(1)
    {
      /*
        if (!record(audioBuf))
            break;
      */
      if(fileind+BUF_SIZE < filesize){
	memcpy(audioBuf, src+fileind, BUF_SIZE);
	fileind+=BUF_SIZE;
      }else{
	fileind=0;
	memcpy(audioBuf, src+fileind, BUF_SIZE);
      }

      /* Interleave the audioBuf */
      int ptr=0;
      while(ptr < BUF_SIZE)
	for(int i=0; i<INTLV_DEGREE; i++, ptr++)
	  intLvBuf[i][ptr/INTLV_DEGREE] = audioBuf[ptr];
      
      for(int i=0;i<INTLV_DEGREE;i++)
	encode(intLvBuf[i], encodeBuf[i]);
      
        //coded stream in encodeBuf
        //sending out to remote
      
      for(int i=0;i<INTLV_DEGREE;i++) {
	bcopy(encodeBuf[i], pckt.data, INTLV_COMP_SIZE);
	pckt.tupleid = id++;
	
	if (send(s, &pckt, sizeof(Pkt), 0) <sizeof(Pkt)) {
	  perror("sendto");
	  exit(1);
	}
	usleep(SLEEP_TIME/INTLV_DEGREE); //this again needs to be synched	
      }
      
      num++;
      if(num%100 == 0)
	printf("Sent %d packets.\n", num);
      
    }
    
    //close_audio();
    close(s);
}
