#include "public.h"
#include "common.h"
#include "g72x.h"
#include "audio.h"

// include files for socket communication
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <cstring>
#include <errno.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/socket.h>

bool play(AUDIOFMT* decodeBuf)
{
    if (write(audio_fd, (char*)&decodeBuf[0], BUF_SIZE_BYTE)!= BUF_SIZE_BYTE)
        return false;
            
    return true;
}

int unpack_input(unsigned char *code, int bits, unsigned char* encodeBuf)
{   
    static unsigned int in_buffer = 0;
    static int      in_bits = 0;
    unsigned char       in_byte;

    static int i=0;

    if (in_bits < bits)
    {   
        in_byte=encodeBuf[i];
        i=(i+1)%COMP_SIZE;
        in_buffer |= (in_byte << in_bits);
        in_bits += 8;
    }
    *code = in_buffer & ((1 << bits) - 1);
    in_buffer >>= bits;
    in_bits -= bits;
    return (in_bits > 0);
}


void decode(AUDIOFMT* decodeBuf, unsigned char* encodeBuf)
{
    unsigned char       code;
    signed short sample;

    int i;

    static int decodeFirst=1;
    static struct g72x_state  destate;

    if (decodeFirst==1)
    {
 	// This segment should be called for the first frame only.
	// Subsequent frames should bypass this part of the loop.
        g72x_init_state(&destate);
        decodeFirst=0;
    }

    for (i=0;i<BUF_SIZE;i++)
    {
        unpack_input(&code,4, encodeBuf);
        //decodeBuf[i]=(short)g721_decoder(code,AUDIO_ENCODING_LINEAR,&destate);
        decodeBuf[i]=(short)g721_decoder(code,AUDIO_ENCODING_ULAW,&destate);
    }

}

void main(int argc, char *argv[])
{
    AUDIOFMT decodeBuf[BUF_SIZE];
    unsigned char encodeBuf[COMP_SIZE];
    unsigned char orgBuf[BUF_SIZE];
    double num=0.0, den=0.0, snr=0.0;
    char src[1040000];
   
    FILE *fstart = fopen("orgstream.count", "rb"); //the orginal stream
    assert(fstart != NULL);
  int fileind=0, filesize=0;
  int tmp=0;
  while((tmp=fread(src+fileind, sizeof(char), BUF_SIZE, fstart)) == BUF_SIZE)
    fileind+=tmp;
  fileind+=tmp;

  //printf("Read in %d bytes.\n", fileind);
  filesize=fileind;
  fileind=0;
  fclose(fstart);//done
 
    //initialize the socket to hear on...
    struct hostent *he;
    int s, conn_s;
    int datalen;
    struct sockaddr_in ls;
    
    //initialize the TCP socket to get data from recver
    if((s = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
      perror("socket creation failed");
      exit(1);
    }

    bzero(&ls, sizeof(struct sockaddr_in));
   
    ls.sin_family = AF_INET;
    ls.sin_port = htons(pPORT);
    inet_aton(ADDR_P, (struct in_addr *)&ls.sin_addr);

    //connect to the local receiver
    if(bind(s, (struct sockaddr *)&ls, sizeof(struct sockaddr_in)) == -1) {
      perror("bind recv failed: ");
      exit(1);
    }
    
    //listen and accept a connect
    if(listen(s, 5) == -1){
      perror("listen recv: ");
      exit(2);
    }
    int ln=sizeof(struct sockaddr_in);
    if((conn_s=accept(s, (struct sockaddr *)&ls, (socklen_t *)&ln)) == -1){
      perror("accept recv: ");
      exit(3);
    }

    //read off conn_s
    
    //open_audio();

    int i=0;
    struct timeval tm;
    int sec=0, usec=0;
    double jitter=0.0;
    double delay=0.0, curr_delay=0.0;
    int sec1=0, usec1=0;
    double avg_lateness=0.0;

    while(1)
    {
      for(int datalen=0, fraglen=0; datalen<COMP_SIZE; ){
	if((fraglen=recv(conn_s, encodeBuf+datalen, COMP_SIZE - datalen, 0)) == -1){
	  perror("recv from forwarded: ");
	  exit(4);
	}
	datalen+=fraglen;
      }
      
      i++;
      //printf("Packet received: %d\n", i++);
      decode(decodeBuf, encodeBuf);
      
      gettimeofday(&tm, NULL);
      if(sec!=0){
	  curr_delay = (tm.tv_sec + tm.tv_usec/1000000.0) - (sec + usec/1000000.0);
	  if(delay != 0){
	  double dd = (curr_delay - delay);
	  if(dd < 0)
		    dd = 0;

	  jitter = (i*jitter + dd)/(i+1);
	}
	
	delay = ((i)*delay + curr_delay)/(i+1);
	
      }
	  // see how much past the deadline this is
	  
	sec = tm.tv_sec;
	usec = tm.tv_usec;

      
      //snr...
      if(fileind+BUF_SIZE < filesize){
	memcpy(orgBuf, src+fileind, BUF_SIZE);
	fileind+=BUF_SIZE;
      }else{
	fileind=0;
	memcpy(orgBuf, src+fileind, BUF_SIZE);
      }
     
      
      for(int z=0; z<BUF_SIZE; z++) {
	      num += (orgBuf[z]*orgBuf[z]);
	      den += (orgBuf[z] - decodeBuf[z])*(orgBuf[z] - decodeBuf[z]);
      }
      snr = 10*(log10(num/den));

     //printf("pkt = %d, jitter = %lf, delay = %lf, snr=%lf\n", i-1, jitter, delay,snr);
     printf("%d %lf %lf %lf\n", i-1, jitter, delay,snr);
     fflush(stdout);
/* 
	if (!play(decodeBuf))
	break;
*/
    }
    snr = 10*(log10(num/den));

    printf("%lf\n", snr);
    //close_audio();
}
